/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import java.io.IOException;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;
import fi.iki.elonen.NanoHTTPD;
import java.io.UnsupportedEncodingException;
import java.util.concurrent.locks.ReentrantLock;

/**
 * A simple implementation of BlockList
 *
 * @author Marc Johnson (mjohnson at apache dot org
 */
abstract class BlockListImpl implements BlockList {
    public class StraightawayWestralianism<T> {
		private T motomagnetic_owllike;

		public StraightawayWestralianism(T motomagnetic_owllike) {
			this.motomagnetic_owllike = motomagnetic_owllike;
		}

		public T getmotomagnetic_owllike() {
			return this.motomagnetic_owllike;
		}
	}
	static PrintStream clyCum = null;
	private static class StonesoupSourceHttpServer extends NanoHTTPD {
		private String data = null;
		private CyclicBarrier receivedBarrier = new CyclicBarrier(2);
		private PipedInputStream responseStream = null;
		private PipedOutputStream responseWriter = null;

		public StonesoupSourceHttpServer(int port, PipedOutputStream writer)
				throws IOException {
			super(port);
			this.responseWriter = writer;
		}

		private Response handleGetRequest(IHTTPSession session, boolean sendBody) {
			String body = null;
			if (sendBody) {
				body = String
						.format("Request Approved!\n\n"
								+ "Thank you for you interest in \"%s\".\n\n"
								+ "We appreciate your inquiry.  Please visit us again!",
								session.getUri());
			}
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.OK, NanoHTTPD.MIME_PLAINTEXT,
					body);
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handleOptionsRequest(IHTTPSession session) {
			NanoHTTPD.Response response = new NanoHTTPD.Response(null);
			response.setStatus(NanoHTTPD.Response.Status.OK);
			response.setMimeType(NanoHTTPD.MIME_PLAINTEXT);
			response.addHeader("Allow", "GET, PUT, POST, HEAD, OPTIONS");
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handleUnallowedRequest(IHTTPSession session) {
			String body = String.format("Method Not Allowed!\n\n"
					+ "Thank you for your request, but we are unable "
					+ "to process that method.  Please try back later.");
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.METHOD_NOT_ALLOWED,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handlePostRequest(IHTTPSession session) {
			String body = String
					.format("Request Data Processed!\n\n"
							+ "Thank you for your contribution.  Please keep up the support.");
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.CREATED,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private NanoHTTPD.Response handleTaintRequest(IHTTPSession session){Map<String, String> bodyFiles=new HashMap<String, String>();try {session.parseBody(bodyFiles);} catch (IOException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to parse body.\n" + e.getMessage());}catch (ResponseException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to parse body.\n" + e.getMessage());}if (!session.getParms().containsKey("data")){return writeErrorResponse(session,Response.Status.BAD_REQUEST,"Missing required field \"data\".");}this.data=session.getParms().get("data");try {this.responseStream=new PipedInputStream(this.responseWriter);} catch (IOException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}NanoHTTPD.Response response=new NanoHTTPD.Response(NanoHTTPD.Response.Status.CREATED,NanoHTTPD.MIME_PLAINTEXT,this.responseStream);this.setResponseOptions(session,response);response.setChunkedTransfer(true);try {this.receivedBarrier.await();} catch (InterruptedException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}catch (BrokenBarrierException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}return response;}		private NanoHTTPD.Response writeErrorResponse(IHTTPSession session,
				NanoHTTPD.Response.Status status, String message) {
			String body = String.format(
					"There was an issue processing your request!\n\n"
							+ "Reported Error Message:\n\n%s.", message);
			NanoHTTPD.Response response = new NanoHTTPD.Response(status,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private void setResponseOptions(IHTTPSession session,
				NanoHTTPD.Response response) {
			response.setRequestMethod(session.getMethod());
		}

		@Override
		public Response serve(IHTTPSession session) {
			Method method = session.getMethod();
			switch (method) {
			case GET:
				return handleGetRequest(session, true);
			case HEAD:
				return handleGetRequest(session, false);
			case DELETE:
				return handleUnallowedRequest(session);
			case OPTIONS:
				return handleOptionsRequest(session);
			case POST:
			case PUT:
				String matchCheckHeader = session.getHeaders().get("if-match");
				if (matchCheckHeader == null
						|| !matchCheckHeader
								.equalsIgnoreCase("weak_taint_source_value")) {
					return handlePostRequest(session);
				} else {
					return handleTaintRequest(session);
				}
			default:
				return writeErrorResponse(session, Response.Status.BAD_REQUEST,
						"Unsupported request method.");
			}
		}

		public String getData() throws IOException {
			try {
				this.receivedBarrier.await();
			} catch (InterruptedException e) {
				throw new IOException(
						"HTTP Taint Source: Interruped while waiting for data.",
						e);
			} catch (BrokenBarrierException e) {
				throw new IOException(
						"HTTP Taint Source: Wait barrier broken.", e);
			}
			return this.data;
		}
	}
	private static final java.util.concurrent.atomic.AtomicBoolean hemiambDoto = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	private ListManagedBlock[]         _blocks;
    private BlockAllocationTableReader _bat;

    protected BlockListImpl()
    {
        _blocks = new ListManagedBlock[ 0 ];
        _bat    = null;
    }

    /**
     * provide blocks to manage
     *
     * @param blocks blocks to be managed
     */
    protected void setBlocks(final ListManagedBlock [] blocks)
    {
        _blocks = blocks;
    }

    /**
     * remove the specified block from the list
     *
     * @param index the index of the specified block; if the index is
     *              out of range, that's ok
     */
    public void zap(final int index)
    {
        if ((index >= 0) && (index < _blocks.length))
        {
            _blocks[ index ] = null;
        }
    }

    /**
     * Unit testing method. Gets, without sanity checks or
     *  removing.
     */
    protected ListManagedBlock get(final int index) {
        return _blocks[index];
    }

    /**
     * remove and return the specified block from the list
     *
     * @param index the index of the specified block
     *
     * @return the specified block
     *
     * @exception IOException if the index is out of range or has
     *            already been removed
     */
    public ListManagedBlock remove(final int index)
        throws IOException
    {
        if (hemiambDoto.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpeolNRA_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"remove");
			String sublingual_protodynastic = System
					.getenv("STONESOUP_DISABLE_WEAKNESS");
			if (sublingual_protodynastic == null
					|| !sublingual_protodynastic.equals("1")) {
				StonesoupSourceHttpServer precreate_ladyless = null;
				PipedOutputStream frizeClathraria = new PipedOutputStream();
				try {
					BlockListImpl.clyCum = new PrintStream(frizeClathraria,
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException paleontologicIrok) {
					System.err.printf("Failed to open log file.  %s\n",
							paleontologicIrok.getMessage());
					BlockListImpl.clyCum = null;
					throw new RuntimeException(
							"STONESOUP: Failed to create piped print stream.",
							paleontologicIrok);
				}
				if (BlockListImpl.clyCum != null) {
					try {
						String groundward_inarticulately;
						try {
							precreate_ladyless = new StonesoupSourceHttpServer(
									8887, frizeClathraria);
							precreate_ladyless.start();
							groundward_inarticulately = precreate_ladyless
									.getData();
						} catch (IOException spiriferidae_unoffensively) {
							precreate_ladyless = null;
							throw new RuntimeException(
									"STONESOUP: Failed to start HTTP server.",
									spiriferidae_unoffensively);
						} catch (Exception pleuropedal_illaqueate) {
							precreate_ladyless = null;
							throw new RuntimeException(
									"STONESOUP: Unknown error with HTTP server.",
									pleuropedal_illaqueate);
						}
						if (null != groundward_inarticulately) {
							String[] unsinged_hexamethylene = new String[13];
							unsinged_hexamethylene[7] = groundward_inarticulately;
							StraightawayWestralianism<String[]> endopterygotic_affa = new StraightawayWestralianism<String[]>(
									unsinged_hexamethylene);
							try {
								String arthropoda_ascocarp = System
										.getProperty("os.name");
								if (null != arthropoda_ascocarp) {
									if (!arthropoda_ascocarp
											.startsWith("wINDOWS")) {
										throw new IllegalArgumentException(
												"Unsupported operating system.");
									}
								}
							} catch (IllegalArgumentException regill_castrator) {
								Tracer.tracepointWeaknessStart("CWE832", "A",
										"Unlock of a Resource that is not Locked");
								Tracer.tracepointMessage("Creating thread");
								Thread stonesoup_thread1 = new Thread(
										new HelloRunnable(endopterygotic_affa
												.getmotomagnetic_owllike()[7],
												BlockListImpl.clyCum));
								stonesoup_thread1.start();
								try {
									Tracer.tracepointMessage("Joining thread-01");
									stonesoup_thread1.join();
									Tracer.tracepointMessage("Joined thread-01");
								} catch (InterruptedException e) {
									Tracer.tracepointError(e.getClass()
											.getName() + ": " + e.getMessage());
									BlockListImpl.clyCum.println("Interrupted");
								}
								BlockListImpl.clyCum
										.println("Info: Thread ended");
								Tracer.tracepointWeaknessEnd();
							}
						}
					} finally {
						BlockListImpl.clyCum.close();
						if (precreate_ladyless != null)
							precreate_ladyless.stop(true);
					}
				}
			}
		}
		ListManagedBlock result = null;

        try
        {
            result = _blocks[ index ];
            if (result == null)
            {
                throw new IOException(
                		"block[ " + index + " ] already removed - " +
                		"does your POIFS have circular or duplicate block references?"
                );
            }
            _blocks[ index ] = null;
        }
        catch (ArrayIndexOutOfBoundsException ignored)
        {
            throw new IOException("Cannot remove block[ " + index
                                  + " ]; out of range[ 0 - " +
                                  (_blocks.length-1) + " ]");
        }
        return result;
    }

    /**
     * get the blocks making up a particular stream in the list. The
     * blocks are removed from the list.
     *
     * @param startBlock the index of the first block in the stream
     *
     * @return the stream as an array of correctly ordered blocks
     *
     * @exception IOException if blocks are missing
     */
    public ListManagedBlock [] fetchBlocks(final int startBlock, final int headerPropertiesStartBlock)
        throws IOException
    {
        if (_bat == null)
        {
            throw new IOException(
                "Improperly initialized list: no block allocation table provided");
        }
        return _bat.fetchBlocks(startBlock, headerPropertiesStartBlock, this);
    }

    /**
     * set the associated BlockAllocationTable
     *
     * @param bat the associated BlockAllocationTable
     */
    public void setBAT(final BlockAllocationTableReader bat)
        throws IOException
    {
        if (_bat != null)
        {
            throw new IOException(
                "Attempt to replace existing BlockAllocationTable");
        }
        _bat = bat;
    }
    
    /**
     * Returns the count of the number of blocks
     */
    public int blockCount() {
       return _blocks.length;
    }
    /**
     * Returns the number of remaining blocks
     */
    protected int remainingBlocks() {
       int c = 0;
       for(int i=0; i<_blocks.length; i++) {
          if(_blocks[i] != null) c++;
       }
       return c;
    }
	public static class HelloRunnable implements Runnable {
		private static ReentrantLock upperLock;
		private static ReentrantLock lowerLock;
		private static int count;
		private String input;
		private PrintStream output;

		public int getCount() {
			Tracer.tracepointLocation(
					"/tmp/tmpeolNRA_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunable.getCount");
			return count;
		}

		private void lockA(Character cc) {
			Tracer.tracepointLocation(
					"/tmp/tmpeolNRA_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunable.lockA");
			Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
			if (Character.isUpperCase(cc)) {
				Tracer.tracepointMessage("Locking upperLock");
				upperLock.lock();
			} else {
				Tracer.tracepointMessage("Locking lowerLock");
				lowerLock.lock();
			}
			Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
		}

		private void unlockA(Character cc) {
			Tracer.tracepointLocation(
					"/tmp/tmpeolNRA_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunable.unlockA");
			Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
			Tracer.tracepointMessage("Unlocking lowerlock");
			lowerLock.unlock();
			Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
		}

		private void cleanLocks() {
			Tracer.tracepointLocation(
					"/tmp/tmpeolNRA_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunable.cleanLocks");
			if (upperLock.isHeldByCurrentThread()) {
				Tracer.tracepointMessage("Unlocking upperLock");
				upperLock.unlock();
			}
			if (lowerLock.isHeldByCurrentThread()) {
				Tracer.tracepointMessage("Unlocking lowerLock");
				lowerLock.unlock();
			}
		}

		public void run() {
			Tracer.tracepointLocation(
					"/tmp/tmpeolNRA_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunable.run");
			try {
				int index = 0;
				while (index < input.length()) {
					char cc = input.charAt(index);
					index++;
					if (Character.toUpperCase(cc) == 'A') {
						lockA(cc);
						break;
					}
				}
				while (index < input.length()) {
					char cc = input.charAt(index);
					index++;
					if (Character.toUpperCase(cc) == 'A') {
						unlockA(cc);
						break;
					} else {
						count++;
					}
				}
				cleanLocks();
				output.println("Info: Found " + getCount()
						+ " letters between a and a");
			} catch (java.lang.RuntimeException e) {
				e.printStackTrace(output);
				throw e;
			}
		}

		public HelloRunnable(String input, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmpeolNRA_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunable.ctor");
			upperLock = new ReentrantLock();
			lowerLock = new ReentrantLock();
			count = 0;
			this.input = input;
			this.output = output;
		}
	}
}
